/*
 * Title:        sag_internal/c_types.hpp
 * Description:  C type ABI
 * $Copyright (c) 2015-2017 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 * @Version:     $Id: c_types.hpp 309474 2017-06-13 10:13:34Z mben $
 */

#ifndef _SAG_INTERNAL_C_TYPES_HPP_
#define _SAG_INTERNAL_C_TYPES_HPP_

// already in extern "C"

/* **** Forward declaration for compound types **** */
struct sag_underlying_data_t;
struct sag_underlying_map_table_t;
struct sag_underlying_vector_table_t;
struct sag_underlying_buffer_table_t;
struct sag_underlying_custom_data_t;

/* **** single-word types that are members of data_t **** */

/** A vector type */
struct sag_underlying_vector_t
{
	struct sag_underlying_vector_table_t *table;
};

/** A map type */
struct sag_underlying_map_t
{
	/** Table data */
	struct sag_underlying_map_table_t *table;
};

/** Byte buffers have a pointer to the buffer and a length */
struct sag_underlying_buffer_t
{
	struct sag_underlying_buffer_table_t *table;
};

/** Custom data can be stored with function pointers to do deletion and
	copying */
struct sag_underlying_custom_t
{
	struct sag_underlying_custom_data_t *data;
};

/** A descriminated union type.
	See sag_data_tag for exact types of members.
	Only works in C11, GNU99 or C++ compilers.
	Should be 1 64-bit word plus the discriminator in size */
struct sag_underlying_data_t 
{
	union {
		// primitive types
		/** bool union member */
		bool boolean;
		/** integer union member */
		int64_t integer;
		/** floating point union member */
		double fp;
		// strings
		/** const char *union member */
		const char *string;
		// complex types
		/** decimal union member */
		struct sag_underlying_decimal_t decimal;
		/** list union member */
		struct sag_underlying_vector_t list;
		/** map union member */
		struct sag_underlying_map_t map;
		/** custom pointer union member */
		struct sag_underlying_custom_t custom;
		/** byte buffer union member */
		struct sag_underlying_buffer_t buffer;
	}; // note: this is only supported in C11, GNU99 or C++. First to be aligned
	/** The discriminator */
	sag_data_tag tag;
};

/* **** internal types for the compound types **** */

// The underlying table type for vectors
struct sag_underlying_vector_table_t
{
	/** Number of items used in the vector */
	uint64_t count;
	/** Capacity of the vector */
	uint64_t capacity;
	/** vector data */
	struct sag_underlying_data_t data[1];
};

// The underlying table data type for maps
struct sag_underlying_map_table_entry_t
{
	uint64_t hash;
	struct sag_underlying_data_t key;
	struct sag_underlying_data_t value;
};

// The underlying data for maps
struct sag_underlying_map_table_t
{
	/** Optional type name of contained data */
	const char *name;
	/** Number of valid items in the map table */
	uint64_t count;
	/** Total capacity of the map table */
	uint64_t capacity;
	/** Number of holes (deleted items) in the map table */
	uint64_t holes;
	/** Table data */
	struct sag_underlying_map_table_entry_t table[1];
};

/** Underlying table type for buffers */
struct sag_underlying_buffer_table_t
{
	/** Length of buffer */
	uint64_t length;
	/** buffer data */
	uint8_t data[1];
};

/** Underlying custom data */
struct sag_underlying_custom_data_t
{
	/** This is refcounted and shallow copied */
	uint64_t refcount;
	/** Pointer to the underlying object */
	void* custom;
	/** Function pointer to delete the underlying object */
	void (*deleter)(void*);
	/** Function pointer to copy the underlying object */
	void *(*copy)(void*);
};

/* **** Other types **** */

/** A message is a payload and a string/string map of metadata */
struct sag_underlying_message_t
{
	/** Message payload */
	struct sag_underlying_data_t payload;
	/** Message metadata */
	struct sag_underlying_map_t metadata;
};

/** A void* in a typed wrapper representing plugins */
struct sag_plugin_t
{
	/** The plugin */
	void *plugin;
};

/** A void* in a typed wrapper representing a status reporter item */
struct sag_status_item_t
{
	/** The item*/
	void *item;
};

// give them all more convenient names
typedef struct sag_underlying_data_t sag_underlying_data_t;
typedef struct sag_underlying_map_t sag_underlying_map_t;
typedef struct sag_underlying_vector_t sag_underlying_vector_t;
typedef struct sag_underlying_custom_t sag_underlying_custom_t;
typedef struct sag_underlying_buffer_t sag_underlying_buffer_t;
typedef struct sag_underlying_message_t sag_underlying_message_t;
typedef struct sag_plugin_t sag_plugin_t;
typedef struct sag_status_item_t sag_status_item_t;

#endif // _SAG_INTERNAL_C_TYPES_HPP_
